import nose
import pathlib

from common import *  # noqa


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
async def test_abort_stor(loop, client, server, *, tmp_dir):

    async def request_abort():

        await asyncio.sleep(0.1, loop=loop)
        await client.abort()
        future.set_result(True)

    future = asyncio.Future(loop=loop)
    await client.login()
    stream = await client.upload_stream("/test.txt")
    loop.create_task(request_abort())
    while True:

        try:

            for _ in range(5):

                await stream.write(b"-" * 8192)
                await asyncio.sleep(0.1, loop=loop)

            await stream.finish()
            raise Exception("Not aborted")

        except (ConnectionResetError, BrokenPipeError):

            break

    await future
    await client.quit()
    file = (tmp_dir / "test.txt")
    nose.tools.ok_(file.stat().st_size)
    file.unlink()


class FakeSlowPathIO(aioftp.PathIO):

    async def exists(self, path):

        return True

    async def is_file(self, path):

        return True

    async def _open(self, path, *args, **kwargs):

        return path

    async def close(self, *args, **kwargs):

        return

    async def read(self, *args, **kwargs):

        await asyncio.sleep(0.05, loop=self.loop)
        return b"-" * 8192

    def list(self, path):

        # infinite list
        files = tuple(path.glob("*"))

        class Lister(aioftp.AbstractAsyncLister):

            async def __aiter__(self):

                return self

            async def __anext__(self):

                return files[0]

        return Lister(timeout=self.timeout, loop=self.loop)


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)],
                 {"path_io_factory": FakeSlowPathIO}))
@with_connection
@with_tmp_dir("foo")
async def test_abort_retr(loop, client, server, *, tmp_dir):

    async def request_abort():

        await asyncio.sleep(0.1, loop=loop)
        await client.abort()
        future.set_result(True)

    future = asyncio.Future(loop=loop)
    await client.login()
    stream = await client.download_stream("/test.txt")
    loop.create_task(request_abort())
    while True:

        data = await stream.read(8192)
        if not data:

            break

    await future
    await client.quit()


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo"),)],
                 {"path_io_factory": FakeSlowPathIO}))
@expect_codes_in_exception("426")
@with_connection
@with_tmp_dir("foo")
async def test_abort_retr_no_wait(loop, client, server, *, tmp_dir):

    async def request_abort():

        await asyncio.sleep(0.1, loop=loop)
        await client.abort(wait=False)
        future.set_result(True)

    future = asyncio.Future(loop=loop)
    await client.login()
    stream = await client.download_stream("/test.txt")
    loop.create_task(request_abort())
    while True:

        data = await stream.read(8192)
        if not data:

            await stream.finish()
            break

    await future
    await client.quit()


@aioftp_setup()
@with_connection
async def test_nothing_to_abort(loop, client, server):

    await client.login()
    await client.abort()
    await client.quit()


@aioftp_setup(
    server_args=([(aioftp.User(base_path="tests/foo", home_path="/"),)],
                 {"path_io_factory": FakeSlowPathIO}))
@with_connection
@with_tmp_dir("foo")
async def test_mlsd_abort(loop, client, server, *, tmp_dir):

    tmp_file = tmp_dir / "test.txt"
    tmp_file.touch()

    await client.login()
    cwd = await client.get_current_directory()
    nose.tools.eq_(cwd, pathlib.PurePosixPath("/"))

    async for path, info in client.list():

        await client.abort()
        break

    tmp_file.unlink()


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )
    # this test for testing infinite loop of
    # WARNING:asyncio:socket.send() raised exception
    test_abort_stor()
    test_abort_retr()
    test_abort_retr_no_wait()
    test_nothing_to_abort()
    print("done")



import nose
import contextlib

import aioftp.client
import aioftp
from common import *  # noqa
from aioftp.server import MemoryUserManager


@nose.tools.raises(ConnectionResetError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {"idle_timeout": 1}))
@with_connection
async def test_idle_timeout(loop, client, server):

    await asyncio.sleep(2, loop=loop)
    await client.login()


class SlowMemoryPathIO(aioftp.MemoryPathIO):

    @aioftp.pathio.universal_exception
    @aioftp.with_timeout
    async def mkdir(self, path, parents=False, exist_ok=False):

        await asyncio.sleep(10, loop=self.loop)

    @aioftp.pathio.universal_exception
    @aioftp.with_timeout
    async def _open(self, path, mode):

        await asyncio.sleep(10, loop=self.loop)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {
            "path_timeout": 1,
            "path_io_factory": SlowMemoryPathIO,
        }))
@expect_codes_in_exception("451")
@with_connection
async def test_server_path_timeout(loop, client, server):

    await client.login()
    await client.make_directory("foo")


@nose.tools.raises(asyncio.TimeoutError)
@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {}),
    client_args=(
        [],
        {
            "path_timeout": 1,
            "path_io_factory": SlowMemoryPathIO,
        }))
@with_connection
@with_tmp_dir("foo")
async def test_client_path_timeout(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo.txt"
    with f.open("wb") as fout:

        fout.write(b"-" * 1024)

    await client.login()
    try:

        await client.download("foo.txt", "/foo.txt", write_into=True)

    except aioftp.PathIOError as e:

        _, value, _ = e.reason
        raise value

    finally:

        f.unlink()


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {}))
@with_connection
@expect_codes_in_exception("503")
@with_tmp_dir("foo")
async def test_wait_pasv_timeout_fail_short(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo.txt"
    b = b"foobar"

    await client.login()
    await client.command("STOR " + f.name)
    await asyncio.sleep(0.5, loop=loop)
    reader, writer = await client.get_passive_connection("I")

    with contextlib.closing(writer) as writer:

        writer.write(b)
        await writer.drain()

    await client.command(None, "2xx", "1xx")
    await client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b, rb)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {}))
@with_connection
@expect_codes_in_exception("503")
@with_tmp_dir("foo")
async def test_wait_pasv_timeout_fail_long(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo.txt"
    b = b"foobar"

    await client.login()
    await client.command("STOR " + f.name)
    await asyncio.sleep(2, loop=loop)
    reader, writer = await client.get_passive_connection("I")

    with contextlib.closing(writer) as writer:

        writer.write(b)
        await writer.drain()

    await client.command(None, "2xx", "1xx")
    await client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b, rb)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {}))
@with_connection
@with_tmp_dir("foo")
async def test_wait_pasv_timeout_ok(loop, client, server, *, tmp_dir):

    f = tmp_dir / "foo.txt"
    b = b"foobar"

    await client.login()
    await client.command("TYPE I", "200")
    code, info = await client.command("PASV", "227")
    ip, port = client.parse_address_response(info[-1])

    await client.command("STOR " + f.name)
    await asyncio.sleep(0.5, loop=loop)
    reader, writer = await aioftp.client.open_connection(
        ip,
        port,
        loop,
        client.create_connection,
    )

    with contextlib.closing(writer) as writer:

        writer.write(b)
        await writer.drain()

    await client.command(None, "2xx", "1xx")
    await client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b, rb)


@aioftp_setup(
    server_args=(
        [(aioftp.User(base_path="tests/foo", home_path="/"),)],
        {}))
@with_connection
@expect_codes_in_exception("425")
@with_tmp_dir("foo")
async def test_wait_pasv_timeout_ok_but_too_long(loop, client, server, *,
                                                 tmp_dir):

    f = tmp_dir / "foo.txt"
    b = b"foobar"

    await client.login()
    await client.command("TYPE I", "200")
    code, info = await client.command("PASV", "227")
    ip, port = client.parse_address_response(info[-1])

    await client.command("STOR " + f.name)
    await asyncio.sleep(2, loop=loop)

    reader, writer = await aioftp.client.open_connection(
        ip,
        port,
        loop,
        client.create_connection,
    )

    with contextlib.closing(writer) as writer:

        writer.write(b)
        await writer.drain()

    await client.command(None, "2xx", "1xx")
    await client.quit()

    with f.open("rb") as fin:

        rb = fin.read()

    f.unlink()

    nose.tools.eq_(b, rb)


class SlowServer(aioftp.Server):

    async def dispatcher(self, reader, writer):

        await asyncio.sleep(10, loop=self.loop)


@nose.tools.raises(asyncio.TimeoutError)
def test_client_socket_timeout():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)
    server = SlowServer(loop=loop)
    client = aioftp.Client(loop=loop, socket_timeout=1)

    async def coro():
        try:
            await server.start(None, 8888)
            await client.connect("127.0.0.1", 8888)
            await asyncio.sleep(10, loop=loop)
        finally:
            await server.close()

    loop.run_until_complete(coro())


class SlowUserManager(MemoryUserManager):

    @aioftp.with_timeout
    async def get_user(self, login):

        await asyncio.sleep(10, loop=self.loop)


@nose.tools.raises(ConnectionResetError)
def test_user_manager_timeout():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)
    server = aioftp.Server(SlowUserManager(None, timeout=1, loop=loop),
                           loop=loop)
    client = aioftp.Client(loop=loop)

    async def coro():
        try:
            await server.start(None, 8888)
            await client.connect("127.0.0.1", 8888)
            await client.login()
        finally:
            await server.close()

    loop.run_until_complete(coro())


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )

    test_server_path_timeout()
    # test_wait_pasv_timeout_ok_but_too_long()
    print("done")



import functools
import pathlib
import asyncio

import nose

from common import *  # noqa


def test_parse_directory_response():

    parse = functools.partial(aioftp.Client.parse_directory_response, None)

    nose.tools.eq_(
        parse('foo "baz "" test nop" """""fdfs """'),
        pathlib.PurePosixPath('baz " test nop'),
    )


def test_connection_del_future():

    loop = asyncio.new_event_loop()
    c = aioftp.Connection(loop=loop)
    c.foo = "bar"
    del c.future.foo


@nose.tools.raises(AttributeError)
def test_connection_not_in_storage():

    loop = asyncio.new_event_loop()
    c = aioftp.Connection(loop=loop)
    getattr(c, "foo")


@nose.tools.raises(ValueError)
def test_available_connections_too_much_acquires():

    ac = aioftp.AvailableConnections(3)
    ac.acquire()
    ac.acquire()
    ac.acquire()
    ac.acquire()


@nose.tools.raises(ValueError)
def test_available_connections_too_much_releases():

    ac = aioftp.AvailableConnections(3)
    ac.acquire()
    ac.release()
    ac.release()



import asyncio
import contextlib

import nose

import aioftp
from common import *


class MyClient(aioftp.Client):

    async def collect(self, count):

        collected = []
        async with self.get_stream("COLL " + str(count), "1xx") as stream:

            async for line in stream.iter_by_line():

                collected.append(line)

        return collected


class MyServer(aioftp.Server):

    @aioftp.ConnectionConditions(
        aioftp.ConnectionConditions.login_required,
        aioftp.ConnectionConditions.passive_server_started)
    async def coll(self, connection, rest):

        @aioftp.ConnectionConditions(
            aioftp.ConnectionConditions.data_connection_made,
            wait=True,
            fail_code="425",
            fail_info="Can't open data connection")
        @aioftp.server.worker
        async def coll_worker(self, connection, rest):

            stream = connection.data_connection
            del connection.data_connection

            async with stream:

                for i in range(count):

                    await stream.write(str.encode(str(i) + "\n"))

            connection.response("200", "coll transfer done")
            return True

        count = int(rest)
        coro = coll_worker(self, connection, rest)
        task = connection.loop.create_task(coro)
        connection.extra_workers.add(task)
        connection.response("150", "coll transfer started")
        return True


def test_stream_iter_by_line():

    async def worker():

        server = MyServer(loop=loop)
        client = MyClient(loop=loop)

        await server.start("127.0.0.1", 8021)
        await client.connect("127.0.0.1", 8021)
        await client.login()

        count = 20
        expect = list(map(lambda i: str.encode(str(i) + "\n"), range(count)))
        collected = await client.collect(count)
        nose.tools.eq_(collected, expect)
        await client.quit()
        client.close()
        await server.close()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)

    loop.run_until_complete(worker())


class CustomException(Exception):

    pass


@aioftp_setup()
@with_connection
async def test_stream_close_without_finish(loop, client, server):

    def fake_finish(*args, **kwargs):

        raise Exception("Finish called")

    await client.login()
    with contextlib.suppress(CustomException):

        async with client.get_stream() as stream:

            stream.finish = fake_finish
            raise CustomException()


@nose.tools.raises(ConnectionRefusedError)
@aioftp_setup()
async def test_no_server(loop, *args, **kwargs):

    async with aioftp.ClientSession("127.0.0.1", loop=loop):

        pass


@aioftp_setup()
@with_connection
async def test_syst_command(loop, client, server):

    await client.login()
    code, info = await client.command("syst", "215")
    nose.tools.eq_(info, [" UNIX Type: L8"])



from common import *  # noqa


@nose.tools.raises(OSError)
@aioftp_setup()
async def test_client_without_server(loop, client, server):

    await client.connect("127.0.0.1", PORT)


@aioftp_setup()
@with_connection
async def test_connection(loop, client, server):

    pass


@aioftp_setup()
@with_connection
async def test_quit(loop, client, server):

    await client.quit()


@aioftp_setup()
@with_connection
async def test_dirty_quit(loop, client, server):

    await asyncio.sleep(0.5, loop=loop)
    client.close()
    await asyncio.sleep(0.5, loop=loop)


@aioftp_setup()
@with_connection
@expect_codes_in_exception("502")
async def test_not_implemented(loop, client, server):

    await client.command("FOOBAR", "2xx", "1xx")


@aioftp_setup()
@with_connection
@expect_codes_in_exception("502")
async def test_type_not_implemented(loop, client, server):

    await client.login()
    await client.get_passive_connection("A")


@nose.tools.timed(2)
@nose.tools.raises(ConnectionResetError, BrokenPipeError)
@aioftp_setup()
@with_connection
async def test_extra_pasv_connection(loop, client, server):

    await client.login()
    r, w = await client.get_passive_connection()
    er, ew = await client.get_passive_connection()
    while True:

        w.write(b"-" * 8192)
        await asyncio.sleep(0.1, loop=loop)
        await w.drain()

    await client.quit()


@nose.tools.timed(2)
@nose.tools.raises(ConnectionError)
@aioftp_setup()
@with_connection
async def test_closing_pasv_connection(loop, client, server):

    await client.login()
    r, w = await client.get_passive_connection()
    host, port = w.transport.get_extra_info("peername")
    nr, nw = await asyncio.open_connection(host, port, loop=loop)
    while True:

        nw.write(b"-" * 100)
        await asyncio.sleep(0.1, loop=loop)
        await nw.drain()

    await client.quit()


@aioftp_setup()
@with_connection
async def test_pasv_connection_ports_not_added(loop, client, server):

    await client.login()
    await client.get_passive_connection()
    await client.quit()

    nose.tools.eq_(server.available_data_ports, None)


@aioftp_setup(server_args=([], {"data_ports": [30000, 30001]}))
@with_connection
async def test_pasv_connection_ports(loop, client, server):

    clients = [aioftp.Client(loop=loop) for _ in range(2)]
    expected_data_ports = [30000, 30001]

    for i, client in enumerate(clients):

        await client.connect("127.0.0.1", PORT)
        await client.login()
        r, w = await client.get_passive_connection()
        host, port = w.transport.get_extra_info("peername")
        nose.tools.eq_(port, expected_data_ports[i])

    for client in clients:

        await client.quit()


@aioftp_setup(server_args=([], {"data_ports": []}))
@with_connection
async def test_data_ports_remains_empty(loop, client, server):

    await client.login()
    await client.quit()

    nose.tools.eq_(server.available_data_ports.qsize(), 0)


@aioftp_setup(server_args=([], {"data_ports": [30000]}))
@with_connection
async def test_pasv_connection_port_reused(loop, client, server):

    clients = [aioftp.Client(loop=loop) for _ in range(2)]

    for client in clients:

        await client.connect("127.0.0.1", PORT)
        await client.login()
        r, w = await client.get_passive_connection()
        host, port = w.transport.get_extra_info("peername")
        nose.tools.eq_(port, 30000)
        await client.quit()


@aioftp_setup(server_args=([], {"data_ports": []}))
@expect_codes_in_exception("421")
@with_connection
async def test_pasv_connection_no_free_port(loop, client, server):

    await client.login()
    await client.get_passive_connection()


@aioftp_setup(server_args=([], {"data_ports": [30000, 30001]}))
@with_connection
async def test_pasv_connection_busy_port(loop, client, server):

    async def handle(reader, writer):
        writer.close()

    conflicting_server = await asyncio.start_server(
        handle,
        host=server.server_host,
        port=30000,
        loop=loop
    )

    await client.login()
    r, w = await client.get_passive_connection()
    host, port = w.transport.get_extra_info("peername")
    nose.tools.eq_(port, 30001)
    await client.quit()

    conflicting_server.close()
    await conflicting_server.wait_closed()


@aioftp_setup(server_args=([], {"data_ports": [30000]}))
@expect_codes_in_exception("421")
@with_connection
async def test_pasv_connection_busy_port2(loop, client, server):

    async def handle(reader, writer):
        writer.close()

    conflicting_server = await asyncio.start_server(
        handle,
        host=server.server_host,
        port=30000,
        loop=loop
    )

    await client.login()
    try:
        await client.get_passive_connection()
    finally:
        conflicting_server.close()
        await conflicting_server.wait_closed()


@aioftp_setup(server_args=([], {"data_ports": [30000, 30001]}))
@expect_codes_in_exception("421")
@with_connection
async def test_pasv_connection_busy_port3(loop, client, server):

    async def handle(reader, writer):
        writer.close()

    conflicting_server1 = await asyncio.start_server(
        handle,
        host=server.server_host,
        port=30000,
        loop=loop
    )

    conflicting_server2 = await asyncio.start_server(
        handle,
        host=server.server_host,
        port=30001,
        loop=loop
    )

    await client.login()
    try:
        await client.get_passive_connection()
    finally:
        conflicting_server1.close()
        conflicting_server2.close()
        await conflicting_server1.wait_closed()
        await conflicting_server2.wait_closed()


@nose.tools.timed(2)
@nose.tools.raises(ConnectionResetError)
@aioftp_setup()
@with_connection
async def test_server_shutdown(loop, client, server):

    async def close_server():

        await asyncio.sleep(1, loop=loop)
        await server.close()

    await client.login()
    loop.create_task(close_server())
    while True:

        await client.list()
        await asyncio.sleep(0.5, loop=loop)


@aioftp_setup()
async def test_client_zeros_passiv_ip(loop, client, server):

    await server.start(None, PORT)
    await client.connect("127.0.0.1", PORT)

    await client.login()
    r, w = await client.get_passive_connection()
    w.close()

    client.close()
    await server.close()


def test_client_session_context_manager():

    async def test():
        server = aioftp.Server(loop=loop)
        await server.start(None, PORT)
        kw = dict(host="127.0.0.1", port=PORT, loop=loop)
        async with aioftp.ClientSession(**kw) as client:
            async for path in client.list():
                pass
        await server.close()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)
    loop.run_until_complete(test())


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    os.environ["PYTHONASYNCIODEBUG"] = "1"
    logging.basicConfig(
        level=logging.INFO
    )

    test_server_shutdown()
    print("done")



import time

import nose

import aioftp
from common import *  # noqa


@aioftp_setup(
    client_args=(
        [],
        {
            "read_speed_limit": 100 * 1024  # 100 Kib
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_client_read_throttle(loop, client, server, *, tmp_dir):

    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * (3 * 100 * 1024))  # 300 Kib

    start = time.perf_counter()
    await client.login()
    stream = await client.download_stream("tests/foo/foo.txt")
    count = 0
    while True:

        data = await stream.read()
        if not data:

            await stream.finish()
            break

        count += len(data)

    nose.tools.eq_(count, 3 * 100 * 1024)
    nose.tools.ok_(2.5 < (time.perf_counter() - start) < 3.5)
    big_file.unlink()


@aioftp_setup(
    client_args=(
        [],
        {
            "read_speed_limit": 100 * 1024  # 100 Kib
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_client_write_with_read_throttle(loop, client, server, *,
                                               tmp_dir):

    start = time.perf_counter()
    big_file = tmp_dir / "foo.txt"
    await client.login()
    stream = await client.upload_stream("tests/foo/foo.txt")
    for _ in range(3 * 100):  # 300 Kib

        await stream.write(b"-" * 1024)

    await stream.finish()

    with big_file.open() as fin:

        data = fin.read()

    print(len(data), (time.perf_counter() - start))
    nose.tools.eq_(len(data), 3 * 100 * 1024)
    nose.tools.ok_((time.perf_counter() - start) < 0.5)
    big_file.unlink()


@aioftp_setup(
    client_args=(
        [],
        {
            "write_speed_limit": 100 * 1024  # 100 Kib
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_client_read_with_write_throttle(loop, client, server, *,
                                               tmp_dir):

    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * (3 * 100 * 1024))  # 300 Kib

    start = time.perf_counter()
    await client.login()
    stream = await client.download_stream("tests/foo/foo.txt")
    count = 0
    while True:

        data = await stream.read()
        if not data:

            await stream.finish()
            break

        count += len(data)

    nose.tools.eq_(count, 3 * 100 * 1024)
    nose.tools.ok_((time.perf_counter() - start) < 0.5)
    big_file.unlink()


@aioftp_setup(
    client_args=(
        [],
        {
            "write_speed_limit": 100 * 1024  # 100 Kib
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_client_write_throttle(loop, client, server, *, tmp_dir):

    start = time.perf_counter()
    big_file = tmp_dir / "foo.txt"
    await client.login()
    stream = await client.upload_stream("tests/foo/foo.txt")
    for _ in range(3 * 100):  # 300 Kib

        await stream.write(b"-" * 1024)

    await stream.finish()

    with big_file.open() as fin:

        data = fin.read()

    nose.tools.eq_(len(data), 3 * 100 * 1024)
    nose.tools.ok_(2.5 < (time.perf_counter() - start) < 3.5)
    big_file.unlink()


@aioftp_setup(
    client_args=(
        [],
        {
            "write_speed_limit": 100 * 1024  # 100 Kib
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_client_write_throttle_changed_after_creation(loop, client,
                                                            server, *,
                                                            tmp_dir):

    nose.tools.eq_(client.throttle.write.limit, 100 * 1024)
    client.throttle.write.limit = 200 * 1024  # 200 Kib

    start = time.perf_counter()
    big_file = tmp_dir / "foo.txt"
    await client.login()
    stream = await client.upload_stream("tests/foo/foo.txt")
    for _ in range(3 * 100):  # 300 Kib

        await stream.write(b"-" * 1024)

    await stream.finish()

    with big_file.open(mode="rb") as fin:

        data = fin.read()

    nose.tools.eq_(len(data), 3 * 100 * 1024)
    nose.tools.ok_(1 < (time.perf_counter() - start) < 2)
    big_file.unlink()


class SlowPathIO(aioftp.PathIO):

    @aioftp.with_timeout
    async def write(self, fout, data):

        timeout = len(data) / (100 * 1024)  # sleep as 100 Kib per second write
        await asyncio.sleep(timeout, loop=self.loop)
        await super().write(fout, data)


@aioftp_setup(
    client_args=(
        [],
        {
            "write_speed_limit": 100 * 1024,  # 100 Kib per second
            "path_io_factory": SlowPathIO,
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_client_write_throttle_with_slow_io(loop, client, server, *,
                                                  tmp_dir):

    start = time.perf_counter()
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 3 * 100 * 1024)

    await client.login()
    await client.download(
        "tests/foo/foo.txt",
        "tests/foo/bar.txt",
        write_into=True
    )

    nose.tools.ok_(2.5 < (time.perf_counter() - start) < 3.5)
    big_file.unlink()
    received = tmp_dir / "bar.txt"
    received.unlink()


@aioftp_setup(
    client_args=(
        [],
        {
            "read_speed_limit": 200 * 1024,  # 200 Kib per second
            "path_io_factory": SlowPathIO,
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_client_read_throttle_with_too_slow_io(loop, client, server, *,
                                                     tmp_dir):

    start = time.perf_counter()
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 3 * 100 * 1024)

    await client.login()
    await client.download(
        "tests/foo/foo.txt",
        "tests/foo/bar.txt",
        write_into=True
    )

    nose.tools.ok_(2.5 < (time.perf_counter() - start) < 3.5)
    big_file.unlink()
    received = tmp_dir / "bar.txt"
    received.unlink()


@aioftp_setup(
    server_args=(
        [],
        {
            "write_speed_limit": 200 * 1024,  # 200 Kib per second
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_server_global_write_throttle_one_user(loop, client, server, *,
                                                     tmp_dir):

    start = time.perf_counter()
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 3 * 100 * 1024)

    await client.login()
    await client.download(
        "tests/foo/foo.txt",
        "tests/foo/bar.txt",
        write_into=True
    )

    nose.tools.ok_(1 < (time.perf_counter() - start) < 2)
    big_file.unlink()
    received = tmp_dir / "bar.txt"
    received.unlink()


@aioftp_setup(
    server_args=(
        [],
        {
            "write_speed_limit": 200 * 1024,  # 200 Kib per second
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_server_global_write_throttle_multi_users(loop, client, server,
                                                        *, tmp_dir):

    async def worker(fname):

        client = aioftp.Client(loop=loop)
        await client.connect("127.0.0.1", PORT)
        await client.login()
        await client.download(
            "tests/foo/foo.txt",
            str.format("tests/foo/{}", fname),
            write_into=True
        )
        await client.quit()

    fnames = ("bar.txt", "baz.txt", "hurr.txt")
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 3 * 100 * 1024)

    start = time.perf_counter()
    await asyncio.wait(map(worker, fnames), loop=loop)

    nose.tools.ok_(4 < (time.perf_counter() - start) < 5)

    big_file.unlink()
    for fname in fnames:

        received = tmp_dir / fname
        received.unlink()


@aioftp_setup(
    server_args=(
        [],
        {
            "read_speed_limit": 200 * 1024,  # 200 Kib per second
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_server_global_read_throttle_multi_users(loop, client, server, *,
                                                       tmp_dir):

    async def worker(fname):

        client = aioftp.Client(loop=loop)
        await client.connect("127.0.0.1", PORT)
        await client.login()
        await client.upload(
            "tests/foo/foo.txt",
            str.format("tests/foo/{}", fname),
            write_into=True
        )
        await client.quit()

    fnames = ("bar.txt", "baz.txt", "hurr.txt")
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 10 * 100 * 1024)

    start = time.perf_counter()
    await asyncio.wait(map(worker, fnames), loop=loop)

    nose.tools.ok_(14.5 < (time.perf_counter() - start) < 15.5)

    big_file.unlink()
    for fname in fnames:

        received = tmp_dir / fname
        received.unlink()


@aioftp_setup(
    server_args=(
        [],
        {
            "write_speed_limit_per_connection": 200 * 1024,  # 200 Kib/s
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_server_per_connection_write_throttle_multi_users(loop, client,
                                                                server, *,
                                                                tmp_dir):

    async def worker(fname):

        client = aioftp.Client(loop=loop)
        await client.connect("127.0.0.1", PORT)
        await client.login()
        await client.download(
            "tests/foo/foo.txt",
            str.format("tests/foo/{}", fname),
            write_into=True
        )
        await client.quit()

    fnames = ("bar.txt", "baz.txt", "hurr.txt")
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 9 * 100 * 1024)

    start = time.perf_counter()
    await asyncio.wait(map(worker, fnames), loop=loop)

    nose.tools.ok_(4 < (time.perf_counter() - start) < 5)

    big_file.unlink()
    for fname in fnames:

        received = tmp_dir / fname
        received.unlink()


@aioftp_setup(
    server_args=(
        [],
        {
            "read_speed_limit_per_connection": 200 * 1024,  # 200 Kib/s
        },
    ))
@with_connection
@with_tmp_dir("foo")
async def test_server_per_connection_read_throttle_multi_users(loop, client,
                                                               server, *,
                                                               tmp_dir):

    async def worker(fname):

        client = aioftp.Client(loop=loop)
        await client.connect("127.0.0.1", PORT)
        await client.login()
        await client.upload(
            "tests/foo/foo.txt",
            str.format("tests/foo/{}", fname),
            write_into=True
        )
        await client.quit()

    fnames = ("bar.txt", "baz.txt", "hurr.txt")
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 9 * 100 * 1024)

    start = time.perf_counter()
    await asyncio.wait(map(worker, fnames), loop=loop)

    nose.tools.ok_(4 < (time.perf_counter() - start) < 5)

    big_file.unlink()
    for fname in fnames:

        received = tmp_dir / fname
        received.unlink()


@aioftp_setup(
    server_args=(
        [[aioftp.User(write_speed_limit_per_connection=200 * 1024)]],
        {},
    ))
@with_connection
@with_tmp_dir("foo")
async def test_server_user_per_connection_write_throttle_multi_users(loop,
                                                                     client,
                                                                     server, *,
                                                                     tmp_dir):

    async def worker(fname):

        client = aioftp.Client(loop=loop)
        await client.connect("127.0.0.1", PORT)
        await client.login()
        await client.download(
            "tests/foo/foo.txt",
            str.format("tests/foo/{}", fname),
            write_into=True
        )
        await client.quit()

    fnames = ("bar.txt", "baz.txt", "hurr.txt")
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 9 * 100 * 1024)

    start = time.perf_counter()
    await asyncio.wait(map(worker, fnames), loop=loop)

    nose.tools.ok_(4 < (time.perf_counter() - start) < 5)

    big_file.unlink()
    for fname in fnames:

        received = tmp_dir / fname
        received.unlink()


@aioftp_setup(
    server_args=(
        [[aioftp.User(read_speed_limit_per_connection=200 * 1024)]],
        {},
    ))
@with_connection
@with_tmp_dir("foo")
async def test_server_user_per_connection_read_throttle_multi_users(loop,
                                                                    client,
                                                                    server, *,
                                                                    tmp_dir):

    async def worker(fname):

        client = aioftp.Client(loop=loop)
        await client.connect("127.0.0.1", PORT)
        await client.login()
        await client.upload(
            "tests/foo/foo.txt",
            str.format("tests/foo/{}", fname),
            write_into=True
        )
        await client.quit()

    fnames = ("bar.txt", "baz.txt", "hurr.txt")
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 9 * 100 * 1024)

    start = time.perf_counter()
    await asyncio.wait(map(worker, fnames), loop=loop)

    nose.tools.ok_(4 < (time.perf_counter() - start) < 5)

    big_file.unlink()
    for fname in fnames:

        received = tmp_dir / fname
        received.unlink()


@aioftp_setup(
    server_args=(
        [[aioftp.User(write_speed_limit=200 * 1024)]],
        {},
    ))
@with_connection
@with_tmp_dir("foo")
async def test_server_user_global_write_throttle_multi_users(loop, client,
                                                             server, *,
                                                             tmp_dir):

    async def worker(fname):

        client = aioftp.Client(loop=loop)
        await client.connect("127.0.0.1", PORT)
        await client.login()
        await client.download(
            "tests/foo/foo.txt",
            str.format("tests/foo/{}", fname),
            write_into=True
        )
        await client.quit()

    fnames = ("bar.txt", "baz.txt", "hurr.txt")
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 3 * 100 * 1024)

    start = time.perf_counter()
    await asyncio.wait(map(worker, fnames), loop=loop)

    nose.tools.ok_(4 < (time.perf_counter() - start) < 5)

    big_file.unlink()
    for fname in fnames:

        received = tmp_dir / fname
        received.unlink()


@aioftp_setup(
    server_args=(
        [[aioftp.User(read_speed_limit=200 * 1024)]],
        {},
    ))
@with_connection
@with_tmp_dir("foo")
async def test_server_user_global_read_throttle_multi_users(loop, client,
                                                            server, *,
                                                            tmp_dir):

    async def worker(fname):

        client = aioftp.Client(loop=loop)
        await client.connect("127.0.0.1", PORT)
        await client.login()
        await client.upload(
            "tests/foo/foo.txt",
            str.format("tests/foo/{}", fname),
            write_into=True
        )
        await client.quit()

    fnames = ("bar.txt", "baz.txt", "hurr.txt")
    big_file = tmp_dir / "foo.txt"
    with big_file.open("wb") as fout:

        fout.write(b"-" * 3 * 100 * 1024)

    start = time.perf_counter()
    await asyncio.wait(map(worker, fnames), loop=loop)

    nose.tools.ok_(4 < (time.perf_counter() - start) < 5)

    big_file.unlink()
    for fname in fnames:

        received = tmp_dir / fname
        received.unlink()


if __name__ == "__main__":

    import logging
    import os

    os.environ["AIOFTP_TESTS"] = "PathIO"
    logging.basicConfig(
        level=logging.INFO
    )

    test_client_write_throttle()
    print("done")



import pathlib

import nose

from common import *  # noqa


async def not_implemented(connection, rest):

    connection.response("502", ":P")
    return True


@aioftp_setup(server_args=([(aioftp.User(base_path="tests/foo"),)], {}))
@with_connection
@with_tmp_dir("foo")
async def test_client_fallback_to_list_at_list(loop, client, server, *, tmp_dir):

    server.mlst = not_implemented
    server.mlsd = not_implemented

    await client.login()
    await client.make_directory("bar")
    async with client.upload_stream("bar/foo.txt") as stream:

        await stream.write(b"foo")

    (path, stat), *_ = files = await client.list()
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.PurePosixPath("bar"))
    nose.tools.eq_(stat["type"], "dir")

    (path, stat), *_ = files = await client.list("bar")
    nose.tools.eq_(len(files), 1)
    nose.tools.eq_(path, pathlib.PurePosixPath("bar/foo.txt"))
    nose.tools.eq_(stat["type"], "file")

    await client.remove("bar")
    files = await client.list()
    nose.tools.eq_(len(files), 0)



import os
import re
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


try:
    version = re.findall(r"""^__version__ = "([^']+)"\r?$""",
                         read(os.path.join("aioftp", "__init__.py")), re.M)[0]
except IndexError:
    raise RuntimeError("Unable to determine version.")


class NoseTestCommand(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pathlib
        import shutil
        import nose

        for path in pathlib.Path("tests").glob("*"):
            if path.is_dir():
                shutil.rmtree(str(path))
        cov = pathlib.Path(".coverage")
        if cov.exists():
            cov.unlink()
        nose.main(argv=["nosetests"])


setup(
    name="aioftp",
    version=version,
    description=("ftp client/server for asyncio"),
    long_description=read("README.rst"),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Internet :: File Transfer Protocol (FTP)",
    ],
    author="pohmelie",
    author_email="multisosnooley@gmail.com",
    url="https://github.com/aio-libs/aioftp",
    license="Apache 2",
    packages=find_packages(),
    install_requires=[],
    tests_require=["nose", "coverage"],
    cmdclass={"test": NoseTestCommand},
    include_package_data=True
)



from common import *  # noqa


@aioftp_setup(
    server_args=([(aioftp.User(
        base_path="tests/foo",
        home_path="/",
        permissions=[aioftp.Permission(writable=False)],
    ),)], {}))
@with_connection
@expect_codes_in_exception("550")
async def test_permission_denied(loop, client, server):

    await client.login()
    await client.make_directory("bar")
    await client.quit()


@aioftp_setup(
    server_args=([(aioftp.User(
        base_path="tests/foo",
        home_path="/",
        permissions=[
            aioftp.Permission("/", writable=False),
            aioftp.Permission("/bar"),
            aioftp.Permission("/foo"),
        ],
    ),)], {}))
@with_connection
@with_tmp_dir("foo")
async def test_permission_overriden(loop, client, server, *, tmp_dir):

    await client.login()
    await client.make_directory("bar")
    await client.remove_directory("bar")
    await client.quit()



import socket

from common import *


@aioftp_setup()
@with_connection(host="::1", family=socket.AF_INET6)
async def test_ipv6_pasv_rejected(loop, client, server):
    await client.login()
    await client.list()
